#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <math.h>

#include "tiledef.h"

#define BRANCH(n)		(((n) << 1) | 0x01)				/* [0, 127] */

#define OP_TILE(n)		(((n) << 2) | 0x02)				/* [0, 63] */
//#define OP_SKIP(n)		(((n) << 3) | 0x04)				/* [0, 31] */
//#define OP_REPEAT(n)		(((n) << 4) | 0x08)				/* [0, 15] */
//#define OP_CTRL(n)		(((n) << 4) | 0x00)				/* [0, 15] */
#define OP_REPEAT(n)		(((n) << 3) | 0x04)				/* [0, 31] */
#define OP_CTRL(n)		(((n) << 3) | 0x00)				/* [0, 31] */

#define OPC_SWAP		1
#define OPC_BG			2
#define OPC_FG			4

#define COLOUR(n)		(((n) << 1) | 0x00)				/* [0, 127] */

#define PAT_NOTE(n)		(((n) << 2) | 0x02)				/* [0, 63] */
#define PAT_INSTR(n)		(((n) << 3) | 0x04)				/* [0, 31] */
#define PAT_DONE		(OP_CTRL(OPC_BG | OPC_FG))

#define SNG_TRACK(n)		(((n) << 1) | 0x00)				/* [0, 96] */
#define SNG_TRANSP(n)		((((n) + 96 + 16) << 1) | 0x00)			/* [-16, 15] */
#define SNG_DONE		(SNG_TRANSP(0))

#define ST_CALL(n)		(((n) << 1) | 0x00)				/* [0, 31] */
#define ST_MAIN(n)		(((n) << 1) | 0x40)				/* [0, 31] */
#define ST_VBL(n)		(((n) << 1) | 0x80)				/* [0, 31] */
#define ST_UNPACK(n)		(((n) << 1) | 0xc0)				/* [0, 31] */
#define ST_PARAM(n)		(((n) << 1) | 0x00)				/* [0, 127] */

#define WIDTH 19
#define HEIGHT 8

#define MAXPIC 16
#define MAXSEQ (WIDTH * HEIGHT * 6 + 1)

#define MAXILEN 64
#define MAXSLEN 256
#define TRACKLEN 24
#define MAXINSTR 32
#define MAXTRACK 256

#define MAXJUMP 128
#define MAXSEGMENT 32
#define MAXFIELD 32

enum {
	N_UNUSED,
	N_LEAF,
	N_BRANCH
};

struct sortable {
	uint16_t	weight;
	uint8_t		value;
};

struct node {
	uint16_t	weight;
	uint16_t	origweight;
	uint8_t		type;
	uint8_t		parent;
	uint8_t		value;	// if leaf
	uint8_t		alt[2];	// if branch
	uint8_t		address;
};

struct hufftable {
	int		len;
	int		addr, root;
	struct node	node[256];
	uint8_t		leaves[256];
};

struct cell {
	uint8_t		tile;
	uint8_t		bg[2];
	uint8_t		fg[2];
};

struct seq {
	uint8_t		sequence[MAXSEQ];
	int		len;
};

struct picture {
	char		name[32];
	struct cell	cell[HEIGHT][WIDTH];
	struct seq	sequence;
};

struct instrument {
	uint8_t		line[MAXILEN];
	int		length;
	int		used;
	int		next;
	int		jumpable;
	int		number;
	int		position;
	int		addr;
};

struct tline {
	uint8_t		note;
	uint8_t		instr;
};

struct track {
	struct tline	line[TRACKLEN];
	uint8_t		number;
	int		used;
	struct seq	sequence;
	int		length;
	struct tchain	*tchain;
	uint16_t	eeaddr;
};

struct sline {
	uint8_t		track[3];
	uint8_t		transp[3];
};

struct song {
	struct sline	line[MAXSLEN];
	int		length;
	struct seq	sequence;
};

struct story {
	struct seq	sequence;
};

struct jump {
	char		*label;
	int		paramflag;
};

struct field {
	int		length;
	int		addr;
	uint8_t		handler[MAXSEGMENT];
};

struct jump jumptable[MAXJUMP];
int njump;

struct field field[MAXFIELD];
int nfield;

uint8_t fieldbytes[256];
int nfieldbytes;

struct story story;

struct hufftable colours, operations;

struct picture picture[MAXPIC];

struct instrument instrument[MAXINSTR];
struct track track[MAXTRACK];
struct song song;

struct tchain *tchain[TRACKLEN + 1];

int ntrack, ninstr;

uint8_t tilemap[64];

uint8_t bitbuf, nbits, xbyte;
int bytecount;

FILE *putbyte_out;
int putbyte_raw;

int odd;

void putbyte(int b) {
	if(putbyte_raw) {
		fputc(b, putbyte_out);
	} else {
		if(xbyte == 0) {
			fprintf(putbyte_out, "\t\t.byte\t");
		} else {
			fprintf(putbyte_out, ", ");
		}
		fprintf(putbyte_out, "0x%02x", b);
		xbyte++;
		if(xbyte == 8) {
			fprintf(putbyte_out, "\n");
			xbyte = 0;
		}
	}
	bytecount++;
	odd ^= 1;
}

void putbit(int b) {
	bitbuf <<= 1;
	bitbuf |= !!b;
	nbits++;
	if(nbits == 8) {
		putbyte(bitbuf);
		bitbuf = 0;
		nbits = 0;
	}
}

void putpad() {
	while(nbits) putbit(0);
	if(xbyte) {
		if(!putbyte_raw) {
			fprintf(putbyte_out, "\n");
		}
		xbyte = 0;
	}
}

int lightest(struct hufftable *ht) {
	int i, argmin = -1;
	uint16_t min = 0xffff;

	for(i = 0; i < ht->len; i++) {
		if(ht->node[i].weight && ht->node[i].weight < min) {
			argmin = i;
			min = ht->node[i].weight;
		}
	}
	return argmin;
}

void initht(struct hufftable *ht) {
	memset(ht, 0, sizeof(struct hufftable));
	memset(ht->leaves, 0xff, 256);
}

void addweight(struct hufftable *ht, uint8_t value) {
	if(ht->leaves[value] == 0xff) {
		ht->leaves[value] = ht->len;
		ht->node[ht->len].type = N_LEAF;
		ht->node[ht->len].value = value;
		ht->node[ht->len].weight = 1;
		ht->len++;
	} else {
		ht->node[ht->leaves[value]].weight++;
	}
}

void enumerate(struct hufftable *ht, uint8_t top) {
	ht->node[top].address = ht->addr++;
	if(ht->node[top].type == N_BRANCH) {
		int diff;

		enumerate(ht, ht->node[top].alt[0]);
		enumerate(ht, ht->node[top].alt[1]);
		diff = ht->node[ht->node[top].alt[1]].address - (ht->node[top].address + 1);
		if(diff > 127) errx(1, "Too large branch (%d) in huffman table", diff);
		ht->node[top].value = BRANCH(diff);
	}
}

int nodecount(struct hufftable *ht, uint8_t top) {
	if(ht->node[top].type == N_BRANCH) {
		return 1
			+ nodecount(ht, ht->node[top].alt[0])
			+ nodecount(ht, ht->node[top].alt[1]);
	} else {
		return 1;
	}
}

void buildtree(struct hufftable *ht) {
	int i, j;

	for(;;) {
		ht->node[ht->len].type = N_BRANCH;
		i = lightest(ht);
		ht->node[ht->len].weight = ht->node[i].weight;
		ht->node[i].origweight = ht->node[i].weight;
		ht->node[i].weight = 0;
		j = lightest(ht);
		if(j == -1) break;
		ht->node[ht->len].weight += ht->node[j].weight;
		ht->node[j].origweight = ht->node[j].weight;
		ht->node[j].weight = 0;
		ht->node[i].parent = ht->len;
		ht->node[j].parent = ht->len;
		if(nodecount(ht, i) < nodecount(ht, j)) {
			ht->node[ht->len].alt[0] = i;
			ht->node[ht->len].alt[1] = j;
		} else {
			ht->node[ht->len].alt[0] = j;
			ht->node[ht->len].alt[1] = i;
		}
		ht->len++;
	}
	ht->node[i].parent = 0xff;

	ht->addr = 0;
	enumerate(ht, i);
}

void printchain(struct hufftable *ht, int n) {
	if(ht->node[n].parent != 0xff) {
		printchain(ht, ht->node[n].parent);
		printf("%d", ht->node[ht->node[n].parent].alt[1] == n);
	}
}

void chain(struct hufftable *ht, int n) {
	if(ht->node[n].parent != 0xff) {
		chain(ht, ht->node[n].parent);
		putbit(ht->node[ht->node[n].parent].alt[1] == n);
	}
}

int chainbits(struct hufftable *ht, int n) {
	if(ht->node[n].parent != 0xff) {
		return 1 + chainbits(ht, ht->node[n].parent);
	}
	return 0;
}

void printht(struct hufftable *ht, int info) {
	int i, j;

	for(i = 0; i < ht->addr; i++) {
		for(j = 0; j < ht->len; j++) {
			if(ht->node[j].address == i) {
				printf("\t\t.byte\t0x%02x", ht->node[j].value);
				if(ht->node[j].type == N_LEAF) {
					uint8_t raw = ht->node[j].value >> 1;
					printf("\t; %02x ", raw);
					switch(info) {
						case 0:
							if(raw & 1) {
								printf("[tile:%02x note:%02x ", raw >> 1, raw >> 1);
							} else if(raw & 2) {
								printf("[rept:%02x inst:%02x ", raw >> 2, raw >> 2);
							} else {
								printf("[ctrl:%02x %4s    ", raw >> 2, (ht->node[j].value == PAT_DONE)? "done" : "");
							}
							if(raw < 96) {
								if(raw < ntrack) {
									printf("track%02d]", raw);
								} else {
									printf("       ]");
								}
							} else if(raw - 96 - 16 == 0) {
								printf("done   ]");
							} else {
								printf("trsp%3d]", raw - 96 - 16);
							}
							break;
						case 1:
							printf("[%02x]", raw + 0x33);
							break;
					}
					printf(" ");
					printchain(ht, j);
				}
				printf("\n");
				odd ^= 1;
			}
		}
	}
}

int seqbytes(struct seq *s) {
	int bits = 0;
	int i;

	for(i = 0; i < s->len; i++) {
		uint8_t value = s->sequence[i];

		bits += chainbits(&operations, operations.leaves[value]);
	}
	return (bits + 7) / 8;
}

void printseq(struct seq *s) {
	int i;

	for(i = 0; i < s->len; i++) {
		uint8_t value = s->sequence[i];

		if(value & 1) {
			chain(&colours, colours.leaves[value & 0xfe]);
		} else {
			chain(&operations, operations.leaves[value]);
		}
	}
	putpad();
}

void printpic(struct picture *p) {
	int i;

	for(i = 0; i < p->sequence.len; i++) {
		uint8_t value = p->sequence.sequence[i];

		if(value & 1) {
			chain(&colours, colours.leaves[value & 0xfe]);
		} else {
			chain(&operations, operations.leaves[value]);
		}
	}
}

void seq(struct seq *s, int kind, uint8_t value) {
	//fprintf(stderr, "%s 0x%02x\n", kind? "COL" : "ACT", value);

	if(s->len >= MAXSEQ) errx(1, "MAXSEQ");
	s->sequence[s->len++] = value | kind;

	if(kind) {
		addweight(&colours, value);
	} else {
		addweight(&operations, value);
	}
}

void seqrepeat(struct picture *p, int repd) {
	while(repd) {
		if(repd > 31) {
			//fprintf(stderr, "rep %d\n", 31);
			seq(&p->sequence, 0, OP_REPEAT(31));
			repd -= 31;
		} else {
			//fprintf(stderr, "rep %d\n", repd);
			seq(&p->sequence, 0, OP_REPEAT(repd));
			repd = 0;
		}
	}
}

void loadsong(char *fname) {
	FILE *f;
	char buf[1024];
	int p[8];

	f = fopen(fname, "r");
	if(!f) err(1, "%s", fname);
	
	while(fgets(buf, sizeof(buf), f)) {
		if(7 == sscanf(buf, "sl %x %x %x %x %x %x %x", &p[0], &p[1], &p[2], &p[3], &p[4], &p[5], &p[6])) {
			song.line[p[0]].track[0] = p[1];
			song.line[p[0]].transp[0] = p[2];
			song.line[p[0]].track[1] = p[3];
			song.line[p[0]].transp[1] = p[4];
			song.line[p[0]].track[2] = p[5];
			song.line[p[0]].transp[2] = p[6];
			if(p[0] >= song.length) song.length = p[0] + 1;
		} else if(4 == sscanf(buf, "tl %x %x %x %x", &p[0], &p[1], &p[2], &p[3])) {
			track[p[0]].line[p[1]].note = p[2];
			track[p[0]].line[p[1]].instr = p[3];
		} else if(3 == sscanf(buf, "il %x %x %x", &p[0], &p[1], &p[2])) {
			instrument[p[0]].line[p[1]] = p[2];
			if(instrument[p[0]].length <= p[1]) instrument[p[0]].length = p[1] + 1;
		}
	}

	fclose(f);
}

int findjump(char *label, int paramflag) {
	int i;

	for(i = 0; i < njump; i++) {
		if(!strcmp(jumptable[i].label, label)) return i;
	}
	jumptable[njump].label = strdup(label);
	jumptable[njump].paramflag = paramflag;
	return njump++;
}

void loadstory(char *fname) {
	FILE *f;
	int currtime = 0;
	char buf[128], str[128];
	int p[2], entry;
	int inlist = 0, inhint = 0;;
	int handler[MAXSEGMENT];
	int height[MAXSEGMENT];
	int nseg;
	int i;
	int line = 0;

	jumptable[0].label = "v_nop";
	jumptable[1].label = "v_grow0";
	njump = 2;

	fieldbytes[0] = findjump("l_end", 0);
	nfieldbytes = 1;

	f = fopen(fname, "r");
	if(!f) err(1, "%s", fname);
	
	while(fgets(buf, sizeof(buf), f)) {
		if(strlen(buf) && buf[strlen(buf) - 1] == 10) buf[strlen(buf) - 1] = 0;
		line++;

		if(2 == sscanf(buf, "%x:%x", &p[0], &p[1])) {
			int diff = p[0] * TRACKLEN + p[1] - currtime;
			currtime += diff;
			while(diff) {
				if(diff >= 96) {
					seq(&story.sequence, 0, ST_CALL(findjump("wait", 0)));
					seq(&story.sequence, 0, ST_PARAM(96));
					diff -= 96;
				} else {
					seq(&story.sequence, 0, ST_CALL(findjump("wait", 0)));
					seq(&story.sequence, 0, ST_PARAM(diff));
					diff = 0;
				}
			}
		} else if(1 == sscanf(buf, " unpack %s", str)) {
			entry = findjump(str, 1);
			seq(&story.sequence, 0, ST_UNPACK(entry));
		} else if(1 == sscanf(buf, " call %s", str)) {
			entry = findjump(str, 0);
			seq(&story.sequence, 0, ST_CALL(entry));
		} else if(1 == sscanf(buf, " main %s", str)) {
			entry = findjump(str, 0);
			seq(&story.sequence, 0, ST_MAIN(entry));
		} else if(1 == sscanf(buf, " vblank %s", str)) {
			entry = findjump(str, 0);
			seq(&story.sequence, 0, ST_VBL(entry));
		} else if(inlist && 2 == sscanf(buf, " %s %d", str, &p[0])) {
			entry = findjump(str, 0);
			if(p[0] & 1) errx(1, "%d: Odd segment heights not allowed", line);
			height[nseg] = (p[0] & 255) >> 1;
			handler[nseg++] = entry;
		} else if(inhint && 1 == sscanf(buf, " %s", str) && strcmp(str, "}")) {
			entry = findjump(str, 0);
			handler[nseg++] = entry;
		} else if(!strcmp(buf, "\tcopper {")) {
			inlist = 1;
			nseg = 0;
		} else if(!strcmp(buf, "\thint {")) {
			inhint = 1;
			nseg = 0;
		} else if((inlist || inhint) && !strcmp(buf, "\t}")) {
			int nmatch = 0;

			// Try to re-use or overlap an earlier field
			for(i = 0; i < nfieldbytes && nmatch < nseg; i++) {
				if(handler[nmatch] == fieldbytes[i]) {
					nmatch++;
				} else {
					// No match here. Backtrack.
					i -= nmatch;
					nmatch = 0;
				}
			}
			//fprintf(stderr, "nmatch %d\n", nmatch);
			entry = i - nmatch;
			while(nfieldbytes < entry + nseg) {
				fieldbytes[nfieldbytes] = handler[nfieldbytes - entry];
				nfieldbytes++;
			}
			if(inlist) {
				seq(&story.sequence, 0, ST_CALL(findjump("loadcopper", 0)));
				seq(&story.sequence, 0, ST_PARAM(entry));
				seq(&story.sequence, 0, ST_PARAM(nseg));
				for(i = 0; i < nseg; i++) {
					seq(&story.sequence, 0, ST_PARAM(height[i]));
				}
			}
			inlist = 0;
			inhint = 0;
		} else if(buf[0] && buf[0] != '#') {
			errx(1, "syntax error in story: '%s'", buf);
		}
	}
	
	fclose(f);
}

uint8_t tilebyte(int t, int y) {
	int p[8];
	int x;

	for(x = 0; x < 8; x++) {
		p[x] = (gfx[t][7 - y][x] == '#');
	}
	return
		((p[0])? 0x80 : 0) |
		((p[1] ^ p[0])? 0x40 : 0) |
		((p[2] ^ p[1])? 0x20 : 0) |
		((p[3] ^ p[1])? 0x10 : 0) |
		((p[4] ^ p[2])? 0x08 : 0) |
		((p[5] ^ p[4])? 0x04 : 0) |
		((p[6] ^ p[5])? 0x02 : 0) |
		((p[7] ^ p[3])? 0x01 : 0);
}

int cmp_weight(const void *p1, const void *p2) {
	const struct sortable *s1 = (const struct sortable *) p1;
	const struct sortable *s2 = (const struct sortable *) p2;
	int diff;

	diff = s2->weight - s1->weight;
	if(!diff) diff = s1->value - s2->value;		// prefer low values; this will ensure consistent behaviour regardless of the qsort implementation
	return diff;
}

void sortvalues(uint8_t *dest, struct hufftable *ht) {
	struct sortable buf[128];
	int i;

	for(i = 0; i < 128; i++) {
		buf[i].value = i * 2;
		if(ht->leaves[i * 2] == 0xff) {
			buf[i].weight = 0;
		} else {
			buf[i].weight = ht->node[ht->leaves[i * 2]].weight;
		}
	}

	qsort(buf, 128, sizeof(struct sortable), cmp_weight);

	for(i = 0; i < 128; i++) {
		dest[i] = buf[i].value;
	}
}

void makeeeprom() {
	struct sortable buf[MAXTRACK];
	int i, n;
	int addr = 0;

	// Decide which tracks go into eeprom.
	// Use a greedy algorithm for now.

	n = 0;
	for(i = 1; i < MAXTRACK; i++) {
		if(track[i].used) {
			buf[n].value = i;
			buf[n].weight = seqbytes(&track[i].sequence);
			n++;
		}
	}

	qsort(buf, ntrack - 1, sizeof(struct sortable), cmp_weight);

	putbyte_out = fopen("eeprom.raw", "wb");
	if(!putbyte_out) err(1, "eeprom.raw");
	putbyte_raw = 1;

	n = 0;
	for(i = 0; i < ntrack - 1; i++) {
		if(addr + buf[i].weight <= 512) {
			track[buf[i].value].eeaddr = 0x8000 + addr;
			printseq(&track[buf[i].value].sequence);
			addr += buf[i].weight;
			n++;
		}
	}
	fclose(putbyte_out);
	putbyte_out = stdout;
	putbyte_raw = 0;

	fprintf(stderr, "%d tracks moved to EEPROM (%d bytes)\n", n, addr);
}

void makeinclude() {
	FILE *fp;
	int i;

	fp = fopen("generated.i", "w");
	if(!fp) err(1, "generated.i");

	for(i = 0; i < 64; i++) {
		if(gfx[i][8]) {
			fprintf(fp, "#define TILE_%-10s0x%02x\n", gfx[i][8], tilemap[i]);
		}
	}

	fclose(fp);
}

void makeaux() {
	bytecount = 0;

	putbyte_out = fopen("generated-aux.S", "w");
	if(!putbyte_out) err(1, "generated-aux.S");

	fprintf(putbyte_out, "\t\t.global\tstory\n");

	fprintf(putbyte_out, "\nstory:\n");
	printseq(&story.sequence);
	
	putbyte_out = stdout;

	fprintf(stderr, "story: %d\n", bytecount);
	bytecount = 0;
}

int main(int argc, char **argv) {
	int i, j, x, y;
	int flag, addr;
	uint8_t toplist[128];
	struct sortable tilehist[64];

	putbyte_out = stdout;

	initht(&colours);
	initht(&operations);

	for(i = 1; i < argc; i++) {
		FILE *f = fopen(argv[i], "rb");
		if(!f) err(1, "%s", argv[i]);

		snprintf(picture[i].name, sizeof(picture[i].name), "%s", argv[i]);
		if(strchr(picture[i].name, '.')) *strchr(picture[i].name, '.') = 0;

		for(y = 0; y < 8; y++) {
			for(x = 0; x < 19; x++) {
				picture[i].cell[y][x].tile = fgetc(f);
			}
			for(x = 0; x < 19; x++) {
				picture[i].cell[y][x].bg[0] = fgetc(f);
				picture[i].cell[y][x].fg[0] = fgetc(f) ^ picture[i].cell[y][x].bg[0];
			}
			for(x = 0; x < 19; x++) {
				picture[i].cell[y][x].bg[1] = fgetc(f);
				picture[i].cell[y][x].fg[1] = fgetc(f) ^ picture[i].cell[y][x].bg[1];
			}
		}

		fclose(f);
	}

	loadsong("demo.song");
	loadstory("story");

	// All data has been loaded into memory.
	
	// Rearrange the music data to pack as efficiently as possible.

	for(i = 0; i < song.length; i++) {
		for(j = 0; j < 3; j++) {
			track[song.line[i].track[j]].used++;
		}
	}

	for(i = 0; i < MAXTRACK; i++) {
		track[i].number = 0xff;
		if(track[i].used) {
			int lasti;

			track[i].length = 0;
			lasti = 255;
			for(j = 0; j < TRACKLEN; j++) {
				instrument[track[i].line[j].instr].used = 1;
				if(track[i].line[j].note && track[i].line[j].instr) {
					if(track[i].line[j].instr == lasti) {
						track[i].line[j].instr = 0;
					} else {
						lasti = track[i].line[j].instr;
					}
				}
				if(track[i].line[j].note || track[i].line[j].instr) {
					track[i].length = j + 1;
				}
			}
		}
	}

	instrument[0].line[0] = 0x4f;
	instrument[0].length = 1;
	instrument[0].jumpable = 1;

	//instrument[7].used = 1;	// hack: this unused instrument is a common suffix of two real istruments

	for(i = 0; i < MAXINSTR; i++) {
		if(instrument[i].length && (instrument[i].line[instrument[i].length - 1] & 0xf0) == 0x00) {
			instrument[i].next = instrument[i].line[--instrument[i].length];
			instrument[instrument[i].next].jumpable = 1;
		}
	}

	for(i = 0; i < MAXINSTR; i++) {
		if(instrument[i].used) {
			if(!instrument[instrument[i].next].used) {
				instrument[instrument[i].next].used = 1;
				i = -1;
			}
		}
	}

	for(i = 0; i < MAXINSTR; i++) {
		if(!instrument[i].used) {
			instrument[i].length = 0;
			instrument[i].next = 0;
		}
	}

	// Optimize the cases where one instrument is a suffix of another.

	flag = 1;
	while(flag) {
		flag = 0;
		for(i = 0; i < MAXINSTR; i++) {
			for(j = 0; j < MAXINSTR; j++) {
				if(i != j
				&& instrument[i].used
				&& instrument[j].used) {
					if(instrument[i].next == instrument[j].next
					&& instrument[i].length <= instrument[j].length) {
						int n;

						for(n = 0; n < instrument[i].length; n++) {
							if(instrument[i].line[n] !=
							instrument[j].line[instrument[j].length - instrument[i].length + n]) {
								break;
							}
						}
						if(n == instrument[i].length) {
							instrument[j].length -= instrument[i].length;
							instrument[j].next = i;
							instrument[i].jumpable = 1;
							flag = 1;
						}
					}
				}
			}
		}
	}

	// Assign new instrument numbers.

	ninstr = 0;
	for(i = 0; i < MAXINSTR; i++) {
		if(instrument[i].used && instrument[i].jumpable) {
			if(ninstr >= 16) errx(1, "Too many jumpable instruments.");
			instrument[i].number = ninstr++;
		}
	}
	for(i = 0; i < MAXINSTR; i++) {
		if(instrument[i].used && !instrument[i].jumpable) {
			instrument[i].number = ninstr++;
		}
	}

	// Re-order instruments to reduce the number of jumps.

	j = 1;		// Position of next instrument, back-to-front
	for(;;) {
		int tail = -1;

		flag = 0;

		// Pick any self-jumping instrument.

		for(i = 0; i < MAXINSTR; i++) {
			if(instrument[i].used
			&& !instrument[i].position
			&& instrument[i].next == i) {
				instrument[i].position = j++;
				tail = i;
				break;
			}
		}

		// If we didn't find any, just pick any jumpable instrument.

		if(tail == -1) {
			for(i = 0; i < MAXINSTR; i++) {
				if(instrument[i].used
				&& !instrument[i].position
				&& instrument[i].jumpable) {
					instrument[i].position = j++;
					tail = i;
					break;
				}
			}
		}

		// If that failed too, exit the loop.

		if(tail == -1) {
			break;
		}

		// Join instruments into a chain.

		while(tail != -1) {
			for(i = 0; i < MAXINSTR; i++) {
				if(instrument[i].used
				&& !instrument[i].position
				&& instrument[i].next == tail) {
					instrument[i].position = j++;
					tail = i;
					break;
				}
			}
			if(i == MAXINSTR) tail = -1;
		}
	}

	for(i = 0; i < MAXINSTR; i++) {
		if(instrument[i].used
		&& !instrument[i].position) {
			instrument[i].position = j++;
		}
	}

	// Re-insert jumps and calculate instrument addresses.

	addr = 0;
	while(--j) {
		for(i = 0; i < MAXINSTR; i++) {
			if(instrument[i].position == j) {
				if(instrument[instrument[i].next].position != j - 1) {
					instrument[i].line[instrument[i].length++] = instrument[instrument[i].next].number;
				}
				instrument[i].addr = addr;
				addr += instrument[i].length;
				break;
			}
		}
	}

	// Generate track data

	for(i = 0; i < MAXTRACK; i++) {
		if(track[i].used) {
			ntrack++;
			for(j = 0; j < track[i].length; j++) {
				struct tline *line = &track[i].line[j];

				if(line->instr) {
					seq(&track[i].sequence, 0, PAT_INSTR(instrument[line->instr].number));
				}
				seq(&track[i].sequence, 0, PAT_NOTE(line->note));
			}
			seq(&track[i].sequence, 0, PAT_DONE);
		}
	}

	// Reorder tracks so that frequent track numbers will coincide with frequent note numbers.
	// This will improve the huffman coding.

	sortvalues(toplist, &operations);

	/*for(i = 0; i < 128; i++) {
		fprintf(stderr, "%02x\n", toplist[i]);
	}*/

	y = 0;
	for(;;) {
		x = 0;
		j = -1;
		for(i = 0; i < MAXTRACK; i++) {
			if(track[i].used && track[i].number == 0xff) {
				if(track[i].used > x) {
					x = track[i].used;
					j = i;
				}
			}
		}
		if(j == -1) {
			break;
		}
		while((toplist[y] >> 1) >= ntrack) y++;	// skip unsuitable numbers
		track[j].number = toplist[y++] >> 1;
	}

	// Generate song data

	for(i = 0; i < song.length; i++) {
		for(j = 0; j < 3; j++) {
			if(song.line[i].transp[j]) {
				seq(&song.sequence, 0, SNG_TRANSP(song.line[i].transp[j]));
			}
			seq(&song.sequence, 0, SNG_TRACK(track[song.line[i].track[j]].number));
		}
	}
	seq(&song.sequence, 0, SNG_DONE);

	// Make a histogram of tiles, skipping those that can be runlength encoded.

	for(i = 0; i < 64; i++) {
		tilehist[i].value = i;
		tilehist[i].weight = 0;
	}

	for(i = 1; i < argc; i++) {
		uint8_t currtile = 255;

		for(y = 0; y < 8; y++) {
			for(x = 0; x < 19; x++) {
				struct cell *cell = &picture[i].cell[y][x];

				if(cell->tile == currtile) {
					// repeat
				} else {
					currtile = cell->tile;
					tilehist[currtile].weight++;
				}
			}
		}
	}

	qsort(tilehist, 64, sizeof(struct sortable), cmp_weight);

	// Reorder tiles to improve huffman coding.

	sortvalues(toplist, &operations);

	/*for(i = 0; i < 128; i++) {
		fprintf(stderr, "%02x\n", toplist[i]);
	}*/

	y = 0;
	for(i = 0; i < 64; i++) {
		while(!(toplist[y] & 0x02)) y++;	// skip unsuitable numbers
		tilemap[tilehist[i].value] = toplist[y++] >> 2;
	}

	for(i = 0; i < 64; i++) {
		//fprintf(stderr, "%02x %2d %02x\n", tilehist[i].value, tilehist[i].weight, tilemap[tilehist[i].value]);
		if(!tilehist[i].weight && !gfx[tilehist[i].value][8]) {
			fprintf(stderr, "Unused tile: %02x\n", tilehist[i].value);
			for(y = 0; y < 8; y++) {
				fprintf(stderr, "%s\n", gfx[tilehist[i].value][y]);
			}
		}
	}

	for(i = 1; i < argc; i++) {
		uint8_t currbg[2] = {0, 0}, currfg[2] = {0, 0}, currtile = 255, repd = 0;

		for(y = 0; y < 8; y++) {
			for(x = 0; x < 19; x++) {
				struct cell *cell = &picture[i].cell[y][x];

				if(cell->tile == currtile
					&& cell->bg[0] == currbg[0]
					&& cell->bg[1] == currbg[1]
					&& ((currtile == 0) || (cell->fg[0] == currfg[0] && cell->fg[1] == currfg[1]))) {

					repd++;
				} else {
					int flags = 0;

					seqrepeat(&picture[i], repd);
					repd = 0;

					if((cell->bg[0] != cell->fg[0] && cell->bg[1] != cell->fg[1])
					&& (currbg[0] != currfg[0] && currbg[1] != currfg[1])) {
						if((cell->bg[0] == currfg[0] && cell->bg[1] == currfg[1])
						|| (cell->fg[0] == currbg[0] && cell->fg[1] == currbg[1] && cell->tile)) {
							uint8_t temp[2];

							flags |= OPC_SWAP;
							temp[0] = currbg[0];
							temp[1] = currbg[1];
							currbg[0] = currfg[0];
							currbg[1] = currfg[1];
							currfg[0] = temp[0];
							currfg[1] = temp[1];
						}
					}

					if(cell->bg[0] != currbg[0] || cell->bg[1] != currbg[1]) {
						currbg[0] = cell->bg[0];
						currbg[1] = cell->bg[1];
						flags |= OPC_BG;
					}
					if(cell->tile && (cell->fg[0] != currfg[0] || cell->fg[1] != currfg[1])) {
						currfg[0] = cell->fg[0];
						currfg[1] = cell->fg[1];
						flags |= OPC_FG;
					}
					if(flags) {
						seq(&picture[i].sequence, 0, OP_CTRL(flags));
						if(flags & OPC_BG) {
							seq(&picture[i].sequence, 1, COLOUR(currbg[0] - 0x33));
							seq(&picture[i].sequence, 1, COLOUR(currbg[1] - 0x33));
						}
						if(flags & OPC_FG) {
							seq(&picture[i].sequence, 1, COLOUR(currfg[0] - 0x33));
							seq(&picture[i].sequence, 1, COLOUR(currfg[1] - 0x33));
						}
					}
					if(currtile == cell->tile) {
						repd++;
					} else {
						currtile = cell->tile;
						seq(&picture[i].sequence, 0, OP_TILE(tilemap[currtile]));
					}
				}
			}
		}
		seqrepeat(&picture[i], repd);
	}

	buildtree(&operations);
	buildtree(&colours);

	// Produce output

	makeeeprom();
	makeinclude();
	makeaux();

	printf("\t\t.global\ttiles\n");
	printf("\t\t.global\tsine\n");
	printf("\t\t.global\tjumptable\n");
	printf("\t\t.global\tfields\n");
	printf("\t\t.global\tophuff\n");
	printf("\t\t.global\tpalhuff\n");
	printf("\t\t.global\tsong\n");
	printf("\t\t.global\ttracks\n");
	printf("\t\t.global\tendtrack\n");
	printf("\t\t.global\tinstruments\n");
	printf("\t\t.global\tfreq\n");
	printf("\t\t.global\tgraphics\n");

	printf("\t\t.global\tf_test\n");

	printf("\t\t.align\t9\n");

	printf("\ntiles:");
	for(y = 0; y < 8; y++) {
		int tile;

		for(tile = 0; tile < 64; tile++) {
			if(tile % 8 == 0) {
				printf("\n\t\t.byte\t");
			} else {
				printf(", ");
			}
			for(i = 0; i < 64; i++) {
				if(tilemap[i] == tile) break;
			}
			printf("0x%02x", tilebyte(i, y));
		}
	}
	printf("\n");

	printf("\nsine:\n");
	j = 0;
	for(i = 0; i < 256; i++) {
		int8_t value = (int) round(127 * sin(i * M_PI * 2 / 256));

		putbyte((256 + value) & 255);
		if((i & 15) == 0) j += (value >> 3) + 16;
	}
	putpad();
	fprintf(stderr, "sinesum = %d\n", j);

	printf("\njumptable:\n");

	for(i = 0; i < njump; i++) {
		if(jumptable[i].paramflag) {
			printf("\t\t.word\t%s\n", jumptable[i].label);
		} else {
			printf("\t\trjmp\t%s\n", jumptable[i].label);
		}
	}

	odd = 0;
	printf("\nfields:\n");

	for(i = 0; i < nfieldbytes; i++) {
		printf("\t\t.byte\t0x%02x\t; %s\n", 0x80 + fieldbytes[i], jumptable[fieldbytes[i]].label);
		addr++;
		odd ^= 1;
	}

	printf("\nfreq:\n");
	for(i = 0; i < 13; i++) {
		double freq = 440 * pow(pow(2, 1.0/12), i + 5 * 12 - 9 - 24);
		uint16_t rounded = (uint16_t) (freq * 65536 / 15625);
		printf("\t\t.word\t0x%04x\n", rounded);
	}

	printf("\ntracks:\n");
	for(i = 0; i < ntrack; i++) {
		for(j = 0; j < MAXTRACK; j++) {
			if(track[j].number == i) break;
		}
		if(track[j].eeaddr) {
			printf("\t\t.word\t0x%04x\t\t; (%02x in song)\n", track[j].eeaddr, j);
		} else {
			printf("\t\t.word\ttrack%d\t\t; (%02x in song)\n", i, j);
		}
	}

	printf("\nophuff:\n");
	printht(&operations, 0);
	printf("\npalhuff:\n");
	printht(&colours, 1);

	bytecount = 0;
	printf("\ngraphics:\n");
	for(i = 1; i < argc; i++) {
		printpic(&picture[i]);
	}
	putpad();
	fprintf(stderr, "gfx: %d\n", bytecount);
	bytecount = 0;

	printf("\nsong:\n");
	printseq(&song.sequence);

	fprintf(stderr, "song: %d\n", bytecount);
	bytecount = 0;

	printf("\nendtrack:");
	for(i = 0; i < MAXTRACK; i++) {
		if(track[i].used && !track[i].eeaddr) {
			printf("\ntrack%d:\n", track[i].number);
			printseq(&track[i].sequence);
		}
	}

	fprintf(stderr, "tracks: %d\n", bytecount);

	printf("\ninstruments:\n");
	printf("\t\t.byte\t0x00\t; Bootstrap\n");
	odd ^= 1;
	for(i = 0; i < ninstr; i++) {
		for(j = 0; j < MAXINSTR; j++) {
			if(instrument[j].used && instrument[j].number == i) break;
		}
		printf("\t\t.byte\t0x%02x\t; Instr %02x (%02x in song)\n", 1 + ninstr + instrument[j].addr, i, j);
		odd ^= 1;
	}
	printf("\n");

	for(x = ninstr; x > 0; x--) {
		for(i = 0; i < MAXINSTR; i++) {
			if(instrument[i].position == x) {
				for(j = 0; j < instrument[i].length; j++) {
					putbyte(instrument[i].line[j]);
				}
				putpad();
				break;
			}
		}
	}

	fprintf(stderr, "Using %d out of 32 jump table entries\n", njump);
	if(njump > 32) errx(1, "njump");

	//if(odd) printf("\n\t\t.byte\t0\t; Pad to word boundary\n");

	printf("\n#include \"tables.S\"\n");

	return 0;
}
